from typing import Any, Iterable, Protocol


class Queue(Protocol):
    def __iter__(__self) -> Any:
        ...

    def __reversed__(__self) -> Any:
        ...

    def __getitem__(__self, __index: int) -> Any:
        ...

    def __setitem__(__self, __index: int, __value: Any) -> None:
        ...

    def __delitem__(__self, __index: int) -> None:
        ...

    def __len__(__self) -> int:
        ...

    def __iadd__(__self, __other: Iterable) -> Any:
        ...

    def append(__self, __value: Any) -> None:
        ...

    def reverse(__self) -> None:
        ...

    def extend(__self, __values: Iterable) -> None:
        ...

    def index(__self, __value: Any) -> int:
        ...

    def pop(__self) -> Any:
        ...

    def popleft(__self) -> Any:
        ...

    def remove(__self, __index: int) -> None:
        ...

    def clear(__self) -> None:
        ...

    def insert(__self, __index: int, __value: Any) -> None:
        ...
